# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2017
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Locate files within a given interval on losc.ligo.org
"""

import warnings

from . import (
    api,
    datasets,
    urls as lurls,
    utils,
)

__all__ = ['get_urls', 'get_event_urls']
__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def get_urls(
        detector, start, end,
        dataset=None,
        tag=None,
        version=None,
        sample_rate=4096,
        format='hdf5',
        host=api.DEFAULT_URL,
):
    """Fetch the metadata from LOSC regarding a given GPS interval

    Parameters
    ----------
    detector : `str`
        the prefix of the relevant GW detector

    start : `int`
        the GPS start time of your query

    end : `int`
        the GPS end time of your query

    dataset : `str`, `None`, optional
        the name of the dataset to query, e.g. ``'GW150914'``

    version : `int`, `None`, optional
        the data-release version for the selected datasets

    sample_rate : `int`, optional, default : ``4096``
        the sampling rate (Hz) of files you want to find

    format : `str`, optional, default: ``'hdf5'``
        the file format (extension) you want to find

    host : `str`, optional
        the URL of the remote LOSC server

    Returns
    -------
    urls : `list` of `str`
        the list of remote file URLs that contain data matching the
        relevant parameters
    """
    start = int(start)
    end = int(end)

    if tag is not None:
        warnings.warn(
            "the `tag` keyword to get_urls is deprecated, GWOSC no longer "
            "releases multiple datasets for events, please use the `dataset` "
            "and `version` keyword arguments to manually select the host "
            "dataset for URLs",
            DeprecationWarning,
        )

    dataset_metadata = [
        ("event",
         lambda x: api.fetch_catalog_event_json(x, host=host),
         ),
        ("run",
         lambda x: api.fetch_run_json(x, detector, start, end, host=host),
         ),
    ]

    for dstype, _get_urls in dataset_metadata:
        if dataset:
            dsets = [dataset]
        else:
            dsets = datasets.find_datasets(
                type=dstype,
                detector=detector,
                segment=(start, end),
                host=host,
            )
        for dataset in dsets:
            # get URL list for this dataset
            urls = _get_urls(dataset)["strain"]

            # match URLs to request
            urls = lurls.match(
                [u['url'] for u in lurls.sieve(
                    urls, detector=detector,
                    sampling_rate=sample_rate, format=format)],
                start, end, version=version)

            # if full span covered, return now
            if utils.full_coverage(urls, (start, end)):
                return urls

    raise ValueError("Cannot find a LOSC dataset for %s covering [%d, %d)"
                     % (detector, start, end))


def get_event_urls(event, format='hdf5', sample_rate=4096, **match):
    meta = api.fetch_catalog_event_json(
        event,
        host=match.pop('host', api.DEFAULT_URL),
    )
    sieve_kw = {k: match.pop(k) for k in list(match.keys()) if
                k not in {'start', 'end', 'tag', 'version'}}
    return lurls.match(
        [u['url'] for u in lurls.sieve(meta['strain'], format=format,
                                       sample_rate=sample_rate, **sieve_kw)],
        **match)



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2018
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwosc.urls`
"""

import pytest

from .. import urls as gwosc_urls

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


@pytest.mark.remote
def test_sieve(gw150914_urls):
    nfiles = len(gw150914_urls)
    sieved = list(gwosc_urls.sieve(gw150914_urls, detector='L1'))
    assert len(sieved) == nfiles // 2
    sieved = list(gwosc_urls.sieve(gw150914_urls, detector='L1',
                                   sample_rate=4096))
    assert len(sieved) == nfiles // 4

    with pytest.raises(KeyError):
        list(gwosc_urls.sieve(gw150914_urls, blah=None))


@pytest.mark.remote
def test_match(gw150914_urls, gw170817_urls):
    urls = [u['url'] for u in gw150914_urls]
    nfiles = len(urls)
    matched = gwosc_urls.match(urls, start=1126259500)  # after 32s files end
    assert len(matched) == nfiles // 2
    for url in matched:
        assert '_R1-' in url


def test_match_tags():
    urls = [
        "/path/to/X-X1_LOSC_C00_4KHZ_R1-0-10.gwf",
        "/path/to/X-X1_LOSC_C01_4KHZ_R1-0-10.gwf",
    ]
    with pytest.raises(ValueError) as exc:
        gwosc_urls.match(urls)
    assert str(exc.value).startswith('multiple LOSC URL tags')

    matched = gwosc_urls.match(urls, tag='C01')
    assert matched == [urls[1]]

    assert not gwosc_urls.match(urls, tag='BLAH')
    assert not gwosc_urls.match(urls, start=1e12)
    assert not gwosc_urls.match(urls, end=0)


# -- local tests


URLS = [
    {'url': 'X-X1_LOSC_TEST_4_V1-0-1.ext',
     'detector': 'X1',
     'sampling_rate': 100},
    {'url': 'Y-Y1_LOSC_TEST_16_V2-1-1.ext',
     'detector': 'Y1',
     'sampling_rate': 200},
    {'url': 'Y-Y1_LOSC_TEST2_16_V2-1-1.ext',
     'detector': 'Y1',
     'sampling_rate': 200},
]


def test_sieve_local():
    assert list(gwosc_urls.sieve(URLS, detector='X1')) == URLS[:1]
    assert list(gwosc_urls.sieve(URLS, sample_rate=200)) == URLS[1:]


def test_match_local():
    urls = [u['url'] for u in URLS]
    with pytest.raises(ValueError) as exc:
        gwosc_urls.match(urls)
    assert str(exc.value).startswith('multiple LOSC URL tags')
    assert gwosc_urls.match(urls, tag="TEST", start=0, end=1) == [urls[0]]
    assert gwosc_urls.match(urls, tag="TEST", version=2) == [urls[1]]
    assert gwosc_urls.match(urls, tag="TEST2", version="R2") == [urls[2]]



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2018
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import json
import re

from six.moves.urllib.request import urlopen
from six.moves.urllib.error import URLError

DEFAULT_URL = "https://www.gw-openscience.org"
MAX_GPS = 99999999999

VERSIONED_EVENT_REGEX = re.compile(r"_[RV]\d+\Z")


# -- JSON handling ------------------------------------------------------------

def fetch_json(url):
    """Fetch JSON data from a remote URL

    Parameters
    ----------
    url : `str`
        the remote URL to fetch

    Returns
    ------
    json : `object`
        the data fetched from ``url`` as parsed by :func:`json.loads`

    See also
    --------
    json.loads
        for details of the JSON parsing

    Examples
    --------
    >>> from gwpy.io.losc import fetch_json
    >>> fetch_json('https://losc.ligo.org/archive/1126257414/1126261510/json/')
    """
    with contextlib.closing(urlopen(url)) as response:
        data = response.read()
        if isinstance(data, bytes):
            data = data.decode('utf-8')
        try:
            return json.loads(data)
        except ValueError as exc:
            exc.args = ("Failed to parse LOSC JSON from %r: %s"
                        % (url, str(exc)),)
            raise


# -- API calls ----------------------------------------------------------------

def _dataset_url(start, end, host=DEFAULT_URL):
    return "{}/archive/{:d}/{:d}/json/".format(host, start, end)


def fetch_dataset_json(gpsstart, gpsend, host=DEFAULT_URL):
    """Returns the JSON metadata for all datasets matching the GPS interval

    Parameters
    ----------
    gpsstart : `int`
        the GPS start of the desired interval

    gpsend : `int`
        the GPS end of the desired interval

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    json
        the JSON data retrieved from LOSC and returned by `json.loads`
    """
    return fetch_json(_dataset_url(gpsstart, gpsend, host=host))


def _event_url(event, host=DEFAULT_URL):
    return "{}/archive/{}/json/".format(host, event)


def fetch_event_json(event, host=DEFAULT_URL):
    """Returns the JSON metadata for the given event

    Parameters
    ----------
    event : `str`
        the name of the event to query

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    json
        the JSON data retrieved from LOSC and returned by `json.loads`
    """
    return fetch_json(_event_url(event, host=host))


def _run_url(run, detector, start, end, host=DEFAULT_URL):
    return "{}/archive/links/{}/{}/{:d}/{:d}/json/".format(
        host, run, detector, start, end,
    )


def fetch_run_json(run, detector, gpsstart=0, gpsend=MAX_GPS,
                   host=DEFAULT_URL):
    """Returns the JSON metadata for the given science run parameters

    Parameters
    ----------
    run : `str`
        the name of the science run, e.g. ``'O1'``

    detector : `str`
        the prefix of the GW detector, e.g. ``'L1'``

    gpsstart : `int`
        the GPS start of the desired interval

    gpsend : `int`
        the GPS end of the desired interval

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    json
        the JSON data retrieved from LOSC and returned by `json.loads`
    """
    return fetch_json(_run_url(run, detector, gpsstart, gpsend, host=host))


def _catalog_url(catalog, host=DEFAULT_URL):
    return "{}/catalog/{}/filelist/".format(
        host, catalog,
    )


def fetch_catalog_json(catalog, host=DEFAULT_URL):
    """"Returns the JSON metadata for the given catalogue

    Parameters
    ----------
    catalog : `str`
        the name of the event catalog, e.g. `GWTC-1-confident`

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    json
        the JSON data retrieved from GWOSC and returnend by
        :func:`json.loads`
    """
    return fetch_json(_catalog_url(catalog, host=host))


def fetch_catalog_event_json(event, version=None, host=DEFAULT_URL):
    """Returns the JSON metadata for the given event in a catalog

    This method queries for all available data-release versions, returning
    the highest available version, unless ``version=<X>`` is specified.

    Parameters
    ----------
    event : `str`
        the name of the event to query

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    version : `int`, `None`, optional
        restrict query to a given data-release version

    Returns
    -------
    json
        the JSON data retrieved from LOSC and returned by `json.loads`
    """
    # if user gave a versioned event (e.g. GW150914_R1), use it directly
    if VERSIONED_EVENT_REGEX.search(event):
        return fetch_event_json(event, host=host)

    # if user specified a version separately, use it
    if version is not None:
        return fetch_event_json("{0}_R{1}".format(event, version), host=host)

    # otherwise find all available versions and return highest
    # its inefficient, but it works..
    versions = _find_catalog_event_versions(event, host=host)
    return fetch_catalog_event_json(event, version=versions[-1], host=host)


def _find_catalog_event_versions(event, host=DEFAULT_URL):
    """List all available versions of a catalogue event
    """
    vers = 1
    found = []
    while True:
        try:
            resp = urlopen(_event_url("{0}_R{1}".format(event, vers),
                                      host=host))
        except URLError:
            break
        else:
            resp.close()
        found.append(vers)
        vers += 1
    if not found:
        raise ValueError("no event datasets found for {!r}".format(event))
    return found



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2018
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwosc.timeline`
"""

try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest

from .. import timeline

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


@pytest.mark.remote
@pytest.mark.parametrize('flag, start, end, result', [
    ('H1_DATA', 1126051217, 1126151217, [
        (1126073529, 1126114861),
        (1126121462, 1126123267),
        (1126123553, 1126126832),
        (1126139205, 1126139266),
        (1126149058, 1126151217),
    ]),
    ('L1_DATA', 1126259446, 1126259478, [
        (1126259446, 1126259478),
    ])
])
def test_get_segments(flag, start, end, result):
    assert timeline.get_segments(flag, start, end) == result


@pytest.mark.remote
def test_timeline_url():
    # check that unknown IFO results in no matches
    with pytest.raises(ValueError):
        timeline.timeline_url('X1', 1126259446, 1126259478)


@mock.patch('gwosc.timeline._find_dataset', return_value='S6')
def test_timeline_url_local(find):
    assert timeline.timeline_url('L1_DATA', 0, 1, host='test') == (
        'test/timeline/segments/json/S6/L1_DATA/0/1/')
    find.assert_called_with(0, 1, 'L1', host='test')



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2018
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for URL handling
"""

import re
from os.path import (basename, splitext)

# LOSC filename re
URL_REGEX = re.compile(
    r"\A((.*/)*(?P<obs>[^/]+)-"
    r"(?P<ifo>[A-Z][0-9])_(L|GW)OSC_"
    r"((?P<tag>[^/]+)_)?"
    r"(?P<samp>\d+(KHZ)?)_"
    r"[RV](?P<version>\d+)-"
    r"(?P<strt>[^/]+)-"
    r"(?P<dur>[^/\.]+)\."
    r"(?P<ext>[^/]+))\Z"
)

VERSION_REGEX = re.compile(r'[RV]\d+')


def sieve(urllist, **match):
    """Sieve a list of LOSC URL metadata dicts based on key, value pairs

    This method simply matches keys from the ``match`` keywords with those
    found in the JSON dicts for a file URL returned by the LOSC API.
    """
    if 'sample_rate' in match and 'sampling_rate' not in match:
        match['sampling_rate'] = match.pop('sample_rate')
    for urlmeta in urllist:
        if any(match[key] != urlmeta[key] for key in match):
            continue
        yield urlmeta


def _match_url(url, start=None, end=None, tag=None, version=None):
    """Match a URL against requested parameters

    Returns
    -------
    None
        if the URL doesn't match the request

    tag, version : `str`, `int`
        if the URL matches the request

    Raises
    ------
    StopIteration
        if the start time of the URL is _after_ the end time of the
        request
    """
    reg = URL_REGEX.match(basename(url)).groupdict()
    if (
            (tag and reg['tag'] != tag) or
            (version and int(reg['version']) != version)
    ):
        return

    # match times
    if end is not None:
        gps = int(reg['strt'])
        if gps >= end:  # too late
            return

    if start is not None:
        gps = int(reg['strt'])
        dur = int(reg['dur'])
        if gps + dur <= start:  # too early
            return

    return reg['tag'], int(reg['version'])


def match(urls, start=None, end=None, tag=None, version=None):
    """Match LOSC URLs for a given [start, end) interval

    Parameters
    ----------
    urls : `list` of `str`
        List of URL paths

    start : `int`
        GPS start time of match interval

    end : `int`
        GPS end time of match interval

    tag : `str`, optional
        URL tag to match, e.g. ``'CLN'``

    version : `int`, optional
        Data release version to match, defaults to highest available
        version

    Returns
    -------
    urls : `list` of `str`
        A sub-list of the input, based on matching, if no URLs are matched,
        the return will be empty ``[]``.
    """
    matched = {}
    matched_tags = set()

    # sort URLs by duration, then start time, then ...
    urls = sorted(
        urls, key=lambda u: splitext(basename(u))[0].split('-')[::-1],
    )

    # format version request
    if VERSION_REGEX.match(str(version)):
        version = version[1:]
    if version is not None:
        version = int(version)

    # loop URLS
    for url in urls:
        m = _match_url(url, start, end, tag=tag, version=version)
        if m is None:
            continue

        mtag, mvers = m
        matched_tags.add(mtag)
        matched.setdefault(mvers, [])
        matched[mvers].append(url)

    # if multiple file tags found, and user didn't specify, error
    if len(matched_tags) > 1:
        tags = ', '.join(map(repr, matched_tags))
        raise ValueError("multiple LOSC URL tags discovered in dataset, "
                         "please select one of: {}".format(tags))

    # extract highest version
    try:
        return matched[max(matched)]
    except ValueError:  # no matched files
        return []



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2017
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Functions for getting information about available datasets

Note: each of these methods deliberately excludes the 'tenyear' dataset.
"""

import multiprocessing.dummy
from operator import itemgetter

from six import raise_from
from six.moves.urllib.error import URLError

from . import (api, utils, catalog as gwosc_catalog)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

IGNORE = {
    "tenyear",
    "history",
}
CATALOGS = [
    "GWTC-1-confident",
    "GWTC-1-marginal",
]


def _match_dataset(targetdetector, detectors, targetsegment, segment):
    """Returns `True` if the dataset metadata matches the target
    """
    if targetdetector not in set(detectors) | {None}:
        return False
    if targetsegment is None or utils.segments_overlap(
            targetsegment, segment):
        return True


def _run_datasets(detector=None, segment=None, host=api.DEFAULT_URL):
    meta = api.fetch_dataset_json(0, api.MAX_GPS, host=host)["runs"]
    for epoch, metadata in meta.items():
        # ignore tenyear, etc...
        if epoch in IGNORE:
            continue
        if _match_dataset(
                detector,
                metadata['detectors'],
                segment,
                (metadata['GPSstart'], metadata['GPSend']),
        ):
            yield epoch


def find_datasets(
        detector=None,
        type=None,
        segment=None,
        host=api.DEFAULT_URL,
):
    """Find datasets available on the given GW open science host

    Parameters
    ----------
    detector : `str`, optional
        prefix of GW detector

    type : `str`, optional
        type of datasets to restrict, one of ``'run'``, ``'event'``, or
        ``'catalog'``

    segment : 2-`tuple` of `int`, `None`, optional
        a GPS ``[start, stop)`` interval to restrict matches to;
        datasets will match if they overlap at any point

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    datasets : `list` of `str`
        the names of all matched datasets, possibly empty

    Examples
    --------
    (Correct as of 2018-03-14)

    >>> from gwosc.datasets import find_datasets
    >>> find_datasets()
    ['GW150914', 'GW151226', 'GW170104', 'GW170608', 'GW170814', 'GW170817',
     'LVT151012', 'O1', 'S5', 'S6']
    >>> find_datasets(detector='V1')
    ['GW170814', 'GW170817']
    >>> find_datasets(type='event')
    ['GW150914', 'GW151226', 'GW170104', 'GW170608', 'GW170814', 'GW170817',
     'LVT151012']
    """
    # get queries
    type = str(type).rstrip("s").lower()
    needruns = type in {"none", "run"}
    needcatalogs = type in {"none", "catalog"}
    needevents = type in {"none", "event"}

    names = set()

    # search for events and datasets
    if needruns:
        names.update(_run_datasets(
            detector=detector,
            host=host,
            segment=segment,
        ))

    # if we're only search for catalogs, unset other kwargs so that
    # _catalog_datasets doesn't use it (it's slow)
    if not needevents:
        detector = None
        segment = None

    # search for catalogs and catalog entries
    if needcatalogs or needevents:
        for catalog in CATALOGS:
            try:
                events = gwosc_catalog.events(
                    catalog,
                    detector=detector,
                    segment=segment,
                    host=host,
                )
            except (URLError, ValueError):  # catalog not found
                continue
            # record catalog itself
            if needcatalogs:
                names.add(catalog)
            # record events
            if needevents:
                names.update(events)

    return sorted(names)


# -- event utilities ----------------------------------------------------------

def event_gps(event, host=api.DEFAULT_URL):
    """Returns the GPS time of an open-data event

    Parameters
    ----------
    event : `str`
        the name of the event to query

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    gps : `float`
        the GPS time of this event

    Examples
    --------
    >>> from gwosc.datasets import event_gps
    >>> event_gps('GW170817')
    1187008882.43
    >>> event_gps('GW123456')
    ValueError: no event dataset found for 'GW123456'
    """
    try:
        return api.fetch_catalog_event_json(event, host=host)['GPS']
    except ValueError as exc:
        raise_from(
            ValueError('no event dataset found for {0!r}'.format(event)),
            exc,
        )


def event_segment(event, detector=None, version=None, host=api.DEFAULT_URL):
    """Returns the GPS ``[start, stop)`` interval covered by an event dataset

    Parameters
    ----------
    event : `str`
        the name of the event

    detector : `str`, optional
        prefix of GW detector

    version : `int`, `None`, optional
        the version of the data release to use,
        defaults to the highest available version

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    start, end : `int`
        the GPS ``[start, end)`` interval covered by this run dataset

    Examples
    --------
    >>> from gwosc.datasets import event_segment
    >>> event_segment("GW150914")
    segment(1126257415, 1126261511)
    """
    data = api.fetch_catalog_event_json(event, host=host, version=version)
    urls = [u["url"] for u in data["strain"] if
            detector in {u["detector"], None}]
    return utils.urllist_extent(urls)


def event_at_gps(gps, host=api.DEFAULT_URL, tol=1):
    """Returns the name of the open-data event matching the GPS time

    This function will return the first event for which
    ``|eventgps - gps| < = tol``.

    Parameters
    ----------
    gps : `float`
        The GPS time to locate

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    tol : `float`, optional
        the search window (in seconds), default: ``1``

    Returns
    -------
    event : `str`
        the name of the matched event

    Raises
    -------
    ValueError
        if no events are found matching the GPS time

    Examples
    --------
    >>> from gwosc.datasets import event_at_gps
    >>> event_at_gps(1187008882)
    'GW170817'
    >>> event_at_gps(1187008882, tol=.1)
    ValueError: no event found within 0.1 seconds of 1187008882
    """
    for event, meta in api.fetch_dataset_json(
            0, api.MAX_GPS, host=host)['events'].items():
        egps = meta['GPStime']
        if abs(egps - gps) <= tol:
            return event
    raise ValueError("no event found within {0} seconds of {1}".format(
        tol, gps))


def event_detectors(event, host=api.DEFAULT_URL, version=None):
    """Returns the `set` of detectors that observed an event

    Parameters
    ----------
    event : `str`
        the name of the event to query

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    version : `int`, `None`, optional
        the data-release version to use, defaults to the highest available
        version

    Returns
    -------
    detectors : `set`
        the set of detectors for which data file URLs are included in
        the data release

    Examples
    --------
    >>> from gwosc.datasets import event_detectors
    >>> event_detectors("GW150914")
    {'H1', 'L1'}
    """
    data = api.fetch_catalog_event_json(event, host=host, version=version)
    return set(url['detector'] for url in data['strain'])


# -- run utilities ------------------------------------------------------------

def run_segment(run, host=api.DEFAULT_URL):
    """Returns the GPS ``[start, stop)`` interval covered by a run dataset

    Parameters
    ----------
    event : `str`
        the name of the event to query

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    start, end : `int`
        the GPS ``[start, end)`` interval covered by this run dataset

    Examples
    --------
    >>> from gwosc.datasets import run_segment
    >>> run_segment('O1')
    segment(1126051217, 1137254417)
    >>> run_segment('Oh dear')
    ValueError: no run dataset found for 'Oh dear'
    """
    try:
        meta = api.fetch_dataset_json(0, api.MAX_GPS, host=host)['runs'][run]
    except KeyError as exc:
        raise ValueError('no run dataset found for {!r}'.format(exc.args[0]))
    return meta['GPSstart'], meta['GPSend']


def run_at_gps(gps, host=api.DEFAULT_URL):
    """Returns the name of the open-data run dataset matching the GPS time

    This function will return the first event for which
    ``start <= gps < end``

    Parameters
    ----------
    gps : `float`
        The GPS time to locate

    host : `str`, optional
        the URL of the LOSC host to query, defaults to losc.ligo.org

    Returns
    -------
    run : `str`
        the name of the matched observing run

    Raises
    -------
    ValueError
        if no datasets are found matching the GPS time

    Examples
    --------
    >>> from gwosc.datasets import run_at_gps
    >>> run_at_gps(1135136350)
    'O1'
    >>> run_at_gps(0)
    ValueError: no run dataset found containing GPS 0
    """
    for run, meta in api.fetch_dataset_json(
            0, api.MAX_GPS, host=host)['runs'].items():
        if run in IGNORE:
            continue
        start, end = meta['GPSstart'], meta['GPSend']
        if start <= gps < end:
            return run
    raise ValueError("no run dataset found containing GPS {0}".format(gps))



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2019
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Catalog-parsing functions
"""

import json

from . import (api, utils)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

CACHE = {}


def clear_cache():
    global CACHE
    CACHE = {}


def download(catalog, host=api.DEFAULT_URL):
    try:
        return CACHE[catalog]
    except KeyError:
        return CACHE.setdefault(
            catalog,
            api.fetch_catalog_json(catalog, host=host)
        )


def _nested_values(data):
    if isinstance(data, dict):
        for key in data:
            for item in _nested_values(data[key]):
                yield item
    else:
        yield data


def datasets(catalog, detector=None, segment=None, host=api.DEFAULT_URL):
    data = download(catalog, host=host)["data"]
    datasets = []
    for event, edata in data.items():
        files = edata["files"]
        revision = files["DataRevisionNum"]
        detectors = set(files["OperatingIFOs"].split())
        if detector not in detectors | {None}:
            continue
        urls = [url for det in detectors for url in _nested_values(files[det])]
        if segment and not (
                urls and
                utils.segments_overlap(segment, utils.urllist_extent(urls))
        ):
            continue
        datasets.append("{0}_{1}".format(event, revision))
    return datasets


def events(catalog, detector=None, segment=None, host=api.DEFAULT_URL):
    return [e.rsplit("_", 1)[0] for e in datasets(
        catalog,
        detector=detector,
        segment=segment,
        host=host,
    )]



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2018
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwosc.datasets`
"""

import re
try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest

from .. import datasets

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


DATASET_JSON = {
    'events': {
        'GW150914': {'GPStime': 12345, 'detectors': ['H1', 'L1']},
        'GW151226': {'GPStime': 12347, 'detectors': ['H1', 'L1']},
    },
    'runs': {
        'S1': {'GPSstart': 0, 'GPSend': 1, 'detectors': ['H1', 'L1', 'V1']},
        'tenyear': None,
    },
}
CATALOG_JSON = {
    'data': {
        'GW150914': {
            'files': {
                'DataRevisionNum': 'R1',
                'OperatingIFOs': "H1 L1",
                'H1': {},
                'L1': {},
            },
        },
    }
}


@pytest.mark.remote
def test_find_datasets():
    sets = datasets.find_datasets()
    for dset in ('S6', 'O1', 'GW150914', 'GW170817'):
        assert dset in sets
    assert 'tenyear' not in sets
    assert 'history' not in sets


@pytest.mark.remote
def test_find_datasets_detector():
    v1sets = datasets.find_datasets('V1')
    assert 'GW170817' in v1sets
    assert 'GW150914' not in v1sets

    assert datasets.find_datasets('X1', type="run") == []


@pytest.mark.remote
def test_find_datasets_type():
    runsets = datasets.find_datasets(type='run')
    assert 'O1' in runsets
    run_regex = re.compile(r'\A[OS]\d+(_\d+KHZ)?(_[RV]\d+)?\Z')
    for dset in runsets:
        assert run_regex.match(dset)

    assert datasets.find_datasets(type='badtype') == []


@pytest.mark.remote
def test_find_datasets_segment():
    sets = datasets.find_datasets(segment=(1126051217, 1137254417))
    assert "GW150914" in sets
    assert "GW170817" not in sets


@mock.patch('gwosc.api.fetch_dataset_json', return_value=DATASET_JSON)
@mock.patch('gwosc.api.fetch_catalog_json', return_value=CATALOG_JSON)
@mock.patch(
    'gwosc.api.fetch_catalog_event_json',
    return_value={
        "strain": [
            {"url": "https://test.com/X-X1-123-456.gwf",
             "detector": "L1",
             },
            {"url": "https://test.com/X-X1-234-567.gwf",
             "detector": "L1",
             },
        ],
    },
)
@mock.patch("gwosc.datasets.CATALOGS", ["GWTC-1-confident"])
def test_find_datasets_local(fcej, fcj, fdj):
    assert datasets.find_datasets() == [
        "GW150914",
        "GWTC-1-confident",
        'S1',
    ]
    assert datasets.find_datasets(detector='V1') == [
        "GWTC-1-confident",
        "S1",
    ]
    assert datasets.find_datasets(type='event') == ['GW150914']
    assert datasets.find_datasets(type='event', detector='V1') == []


@mock.patch(
    "gwosc.catalog.events",
    side_effect=[ValueError, []],
)
def test_find_datasets_catalog_error(_):
    assert datasets.find_datasets(type="catalog") == ["GWTC-1-marginal"]


@pytest.mark.remote
def test_event_gps():
    assert datasets.event_gps('GW170817') == 1187008882.4
    with pytest.raises(ValueError) as exc:
        datasets.event_gps('GW123456')
    assert str(exc.value) == 'no event dataset found for \'GW123456\''


@mock.patch('gwosc.api.fetch_catalog_event_json', return_value={
    'GPS': 12345,
    'something else': None,
})
def test_event_gps_local(fetch):
    assert datasets.event_gps('GW150914') == 12345
    fetch.side_effect = ValueError('test')
    with pytest.raises(ValueError) as exc:
        datasets.event_gps('something')
    assert str(exc.value) == 'no event dataset found for \'something\''


@pytest.mark.remote
def test_event_segment():
    assert datasets.event_segment("GW170817") == (1187006835, 1187010931)


@mock.patch('gwosc.api.fetch_catalog_event_json', return_value={
    "strain": [
        {"url": "https://test.com/X-X1-123-456.gwf", "detector": "X1"},
        {"url": "https://test.com/X-X1-234-567.gwf", "detector": "Y1"},
    ],
})
def test_event_segment_local(fetch):
    assert datasets.event_segment("GW170817") == (123, 801)
    assert datasets.event_segment("GW170817", detector="X1") == (123, 579)


@pytest.mark.remote
def test_event_at_gps():
    assert datasets.event_at_gps(1187008882) == 'GW170817'
    with pytest.raises(ValueError) as exc:
        datasets.event_at_gps(1187008882, tol=.1)
    assert str(exc.value) == 'no event found within 0.1 seconds of 1187008882'


@mock.patch('gwosc.api.fetch_dataset_json', return_value=DATASET_JSON)
def test_event_at_gps_local(fetch):
    assert datasets.event_at_gps(12345) == 'GW150914'
    with pytest.raises(ValueError):
        datasets.event_at_gps(12349)


@pytest.mark.remote
def test_event_detectors():
    assert datasets.event_detectors("GW150914") == {"H1", "L1"}
    assert datasets.event_detectors("GW170814") == {"H1", "L1", "V1"}


@mock.patch("gwosc.api.fetch_catalog_event_json", return_value={
    "strain": [{"detector": "A1"}, {"detector": "A1"}, {"detector": "B1"}],
})
def test_event_detectors_local(fcej):
    assert datasets.event_detectors("test") == {"A1", "B1"}


@pytest.mark.remote
def test_run_segment():
    assert datasets.run_segment('O1') == (1126051217, 1137254417)
    with pytest.raises(ValueError) as exc:
        datasets.run_segment('S7')
    assert str(exc.value) == 'no run dataset found for \'S7\''


@mock.patch('gwosc.api.fetch_dataset_json', return_value=DATASET_JSON)
def test_run_segment_local(fetch):
    assert datasets.run_segment('S1') == (0, 1)
    with pytest.raises(ValueError):
        datasets.run_segment('S2')


@pytest.mark.remote
def test_run_at_gps():
    assert datasets.run_at_gps(1135136350) in {'O1', 'O1_16KHZ'}
    with pytest.raises(ValueError) as exc:
        datasets.run_at_gps(0)
    assert str(exc.value) == 'no run dataset found containing GPS 0'


@mock.patch('gwosc.api.fetch_dataset_json', return_value=DATASET_JSON)
def test_run_at_gps_local(fetch):
    assert datasets.run_at_gps(0) == 'S1'
    with pytest.raises(ValueError):
        datasets.run_at_gps(10)



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2018
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Helpers for tests
"""

import os
try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest

from ..api import fetch_catalog_event_json

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# unset these variables to allow the tests to run with pybuild
os.environ.pop('http_proxy', None)
os.environ.pop('https_proxy', None)

try:
    autouseyield = pytest.yield_fixture(autouse=True)
except AttributeError:  # pytest > 3.0.0
    autouseyield = pytest.fixture(autouse=True)


# force all tests to not rely on the catalog cache
@autouseyield
def clear_catalog_cache():
    with mock.patch.dict("gwosc.catalog.CACHE", clear=True):
        yield


def _event_urls(name):
    return fetch_catalog_event_json(name)['strain']


@pytest.fixture
def gw150914_urls(scope='module'):
    return _event_urls('GW150914')


@pytest.fixture
def gw170817_urls(scope='module'):
    return _event_urls('GW170817')



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2017
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

from operator import itemgetter

from . import (api, datasets, utils)


def get_segments(flag, start, end, host=api.DEFAULT_URL):
    """Return the [start, end) GPS segments for this flag

    Parameters
    ----------
    flag : `str`
        name of flag, e.g. ``'H1_DATA'``

    start : `int`
        the GPS start time of your query

    end : `int`
        the GPS end time of your query

    host : `str`, optional
        the URL of the remote LOSC server

    Returns
    -------
    segments : `list` of `(int, int)` tuples
        a list of `[a, b)` GPS segments
    """
    return list(map(tuple, api.fetch_json(
        timeline_url(flag, start, end, host=host))['segments']))


def timeline_url(flag, start, end, host=api.DEFAULT_URL):
    """Returns the Timeline JSON URL for a flag name and GPS interval
    """
    detector = flag.split('_', 1)[0]
    dataset = _find_dataset(start, end, detector, host=host)
    return '{}/timeline/segments/json/{}/{}/{}/{}/'.format(
        host, dataset, flag, start, end-start)


def _find_dataset(start, end, detector, host=api.DEFAULT_URL):

    # -- utility methods to resolve a (proper) dataset name and its segment

    def _event_version_segment(event):
        versions = api._find_catalog_event_versions(event, host=host)
        last = versions[-1]
        dataset = "{0}_R{1}".format(event, last)
        metadata = api.fetch_event_json(event, host=host)
        return (
            dataset,
            utils.urllist_extent(map(itemgetter("url"), metadata["strain"])),
        )

    def _run_segment(run):
        return run, datasets.run_segment(run, host=host)

    # -- match datasets

    dataset_segments = [
        ("event", _event_version_segment),
        ("run", _run_segment),
    ]

    duration = end-start
    epochs = []

    for dstype, _dataset_segment in dataset_segments:
        dsets = datasets.find_datasets(
            type=dstype,
            detector=detector,
            segment=(start, end),
            host=host,
        )
        for dataset in dsets:
            dataset, segment = _dataset_segment(dataset)
            overlap = min(end, segment[1]) - max(start, segment[0])
            epochs.append((dataset, duration-overlap))

    if not epochs:
        raise ValueError(
            "No datasets found matching [{}, {})".format(start, end))

    return sorted(epochs, key=itemgetter(1, 0))[0][0]



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2018
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwosc.utils`
"""

import pytest

from .. import utils

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def test_url_segment():
    seg = utils.url_segment('X-TEST-123-456.ext')
    assert seg == (123, 579)


@pytest.mark.parametrize('url, segment, result', [
    ('A-B-10-1.ext', (0, 10), False),
    ('A-B-10-1.ext', (5, 11), True),
    ('A-B-10-1.ext', (10, 15), True),
    ('A-B-10-1.ext', (11, 15), False),
])
def test_url_overlaps_segment(url, segment, result):
    assert utils.url_overlaps_segment(url, segment) is result


@pytest.mark.remote
@pytest.mark.parametrize('segment, result', [
    ((1126257415, 1126257415+4096), True),
    ((1126257414, 1126257415+4096), False),
    ((1126257415, 1126257415+4097), False),
])
def test_full_coverage(gw150914_urls, segment, result):
    urls = [u['url'] for u in gw150914_urls]
    assert utils.full_coverage(urls, segment) is result


def test_full_coverage_empty():
    assert utils.full_coverage([], (0, 1)) is False


@pytest.mark.parametrize('seg1, seg2, result', [
    ((10, 11), (0, 10), False),
    ((10, 11), (5, 11), True),
    ((10, 11), (10, 15), True),
    ((10, 11), (11, 15), False),
])
def test_segments_overlap(seg1, seg2, result):
    assert utils.segments_overlap(seg1, seg2) is result



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2018
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwosc.api`
"""

import os.path
import re
from io import BytesIO

try:
    from unittest import mock
    from urllib.error import URLError
except ImportError:  # python < 3
    import mock
    from urllib2 import URLError

import pytest

from .. import api

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def losc_url(path):
    return '{0}/{1}'.format(api.DEFAULT_URL, path)


def check_json_url_list(urllist, keys={'detector', 'format', 'url'}):
    assert isinstance(urllist, list)
    for urld in urllist:
        for key in keys:
            assert key in urld


@pytest.mark.remote
def test_fetch_json():
    url = 'https://www.gw-openscience.org/archive/1126257414/1126261510/json/'
    out = api.fetch_json(url)
    assert isinstance(out, dict)
    assert len(out['events']) == 1
    assert sorted(out['events']['GW150914']['detectors']) == ['H1', 'L1']
    assert set(out['runs'].keys()).issubset(
        {'tenyear', 'O1', 'O1_16KHZ', 'history'},
    )

    # check errors (use legit URL that isn't JSON)
    url2 = os.path.dirname(os.path.dirname(url))
    with pytest.raises(ValueError) as exc:
        api.fetch_json(url2)
    assert str(exc.value).startswith(
        "Failed to parse LOSC JSON from {!r}: ".format(url2))


@mock.patch('gwosc.api.urlopen', return_value=BytesIO(b'{"key": "value"}'))
def test_fetch_json_local(_):
    url = 'anything'
    out = api.fetch_json(url)
    assert isinstance(out, dict)
    assert out['key'] == 'value'


@pytest.mark.remote
def test_fetch_dataset_json():
    start = 934000000
    end = 934100000
    out = api.fetch_dataset_json(start, end)
    assert not out['events']
    assert set(out['runs'].keys()).issubset({'tenyear', 'S6', 'history'})


@mock.patch('gwosc.api.fetch_json')
def test_fetch_dataset_json_local(fetch):
    start = 934000000
    end = 934100000
    api.fetch_dataset_json(start, end)
    fetch.assert_called_with(
        losc_url('archive/{0}/{1}/json/'.format(start, end)))


@pytest.mark.remote
def test_fetch_event_json():
    event = 'GW150914'
    out = api.fetch_event_json(event)
    assert int(out['GPS']) == 1126259462
    assert out['dataset'] == event
    check_json_url_list(out['strain'])


@mock.patch('gwosc.api.fetch_json')
def test_fetch_event_json_local(fetch):
    api.fetch_event_json('GW150914')
    fetch.assert_called_with(losc_url('archive/GW150914/json/'))


@pytest.mark.remote
def test_fetch_run_json():
    run = 'S6'
    detector = 'L1'
    start = 934000000
    end = 934100000
    out = api.fetch_run_json(run, detector, start, end)
    assert out['dataset'] == run
    assert out['GPSstart'] == start
    assert out['GPSend'] == end
    check_json_url_list(out['strain'])


@mock.patch('gwosc.api.fetch_json')
def test_fetch_run_json_local(fetch):
    api.fetch_run_json('S6', 'L1', 934000000, 934100000)
    fetch.assert_called_with(
        losc_url('archive/links/S6/L1/934000000/934100000/json/'))


@pytest.mark.remote
def test_fetch_catalog_json():
    out = api.fetch_catalog_json("GWTC-1-confident")
    assert "distance" in out["parameters"]
    assert out["data"]["GW170817"]["tc"]["best"] == 1187008882.4


@mock.patch("gwosc.api.fetch_json")
def test_fetch_catalog_json_local(fetch):
    api.fetch_catalog_json("GWTC-1-confident")
    fetch.assert_called_with(
        losc_url("catalog/GWTC-1-confident/filelist/")
    )


@pytest.mark.remote
def test_fetch_catalog_event_json():
    event = 'GW150914'
    out = api.fetch_catalog_event_json(event)
    assert int(out["GPS"]) == 1126259462
    assert re.match(r"GW150914_R\d+", out["dataset"])
    check_json_url_list(out["strain"])


@mock.patch("gwosc.api.fetch_event_json")
@mock.patch("gwosc.api.urlopen",
            side_effect=[mock.MagicMock(), URLError("mock")])
def test_fetch_catalog_event_json_local(urlopen, fej):
    api.fetch_catalog_event_json("GW150914")
    assert urlopen.call_count == 2  # two version queries, one JSON queries
    urlopen.assert_any_call(losc_url("archive/GW150914_R1/json/"))
    urlopen.assert_any_call(losc_url("archive/GW150914_R2/json/"))
    fej.assert_called_with("GW150914_R1", host=api.DEFAULT_URL)


@mock.patch("gwosc.api.fetch_event_json")
def test_fetch_catalog_event_json_version(fetch):
    api.fetch_catalog_event_json("GW150914_R1")
    assert fetch.call_count == 1
    fetch.assert_called_with("GW150914_R1", host=api.DEFAULT_URL)

    fetch.reset_mock()

    api.fetch_catalog_event_json("GW150914", version=10)
    assert fetch.call_count == 1
    fetch.assert_called_with("GW150914_R10", host=api.DEFAULT_URL)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWOSC
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Setup the GWOSC package
"""

import sys

from setuptools import (setup, find_packages)

import versioneer

__version__ = versioneer.get_version()

# dependencies
setup_requires = []
if {'test'}.intersection(sys.argv):
    setup_requires.append('pytest_runner')
install_requires = [
    'six>=1.9.0',
]
tests_require = [
    'pytest>=2.7.0',
]
if sys.version_info.major < 3:
    tests_require.extend((
        'more-itertools < 6.0a0',  # pip doesn't pin version properly
        'mock',
    ))
extras_require = {
    'docs': [
        'sphinx',
        'sphinx_rtd_theme',
        'numpydoc',
    ],
}

# get long description from README
with open('README.md', 'rb') as f:
    longdesc = f.read().decode().strip()

# run setup
setup(
    # metadata
    name='gwosc',
    version=__version__,
    description="A python interface to the GW Open Science data archive",
    long_description=longdesc,
    long_description_content_type='text/markdown',
    author='Duncan Macleod',
    author_email='duncan.macleod@ligo.org',
    url='https://github.com/gwpy/gwosc',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
    ],
    # build
    cmdclass=versioneer.get_cmdclass(),
    # content
    packages=find_packages(),
    # dependencies
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
)



# -*- coding: utf-8 -*-
# Copyright Duncan Macleod 2018
#
# This file is part of GWOSC.
#
# GWOSC is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWOSC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWOSC.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwosc.locate`
"""

import os.path
import re

import pytest

from .. import (
    locate,
    urls as gwosc_urls,
    utils,
)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


@pytest.mark.remote
def test_get_urls():
    # test simple fetch for S6 data returns only files within segment
    detector = 'L1'
    start = 934000000
    end = 934100000
    span = (start, end)
    urls = locate.get_urls(detector, start, end)
    for url in urls:
        assert os.path.basename(url).startswith(
            '{}-{}'.format(detector[0], detector))
        assert utils.segments_overlap(
            utils.url_segment(url), span)

    # test fetch for GW170817 data
    assert len(locate.get_urls(
        'L1', 1187007040, 1187009088,
        dataset="GW170817",
    )) == 2

    # assert no hits raises exception
    with pytest.raises(ValueError):  # no data in 1980
        locate.get_urls(detector, 0, 1)
    with pytest.raises(ValueError):  # no Virgo data for S6
        locate.get_urls('V1', start, end)


@pytest.mark.remote
def test_get_urls_deprecated_tag():
    # test `tag` prints a warning
    pytest.deprecated_call(
        locate.get_urls,
        "L1",
        1187007040,
        1187009088,
        tag="TEST",
    )


@pytest.mark.remote
def test_get_event_urls(gw150914_urls):
    # find latest version by brute force
    latestv = sorted(
        gwosc_urls.URL_REGEX.match(
            os.path.basename(u['url'])).groupdict()['version'] for
        u in gw150914_urls)[-1]

    event = 'GW150914'
    urls = locate.get_event_urls(event)
    v_regex = re.compile("_[RV]{}-".format(latestv))
    for url in urls:
        assert url.endswith('.hdf5')  # default format
        assert '_4KHZ_' in url  # default sample rate
        assert v_regex.search(url)  # highest matched version

    urls = locate.get_event_urls(event, version=1)
    v1_regex = re.compile("_[RV]1-")
    for url in urls:
        assert v1_regex.search(url)


@pytest.mark.remote
def test_get_urls_gw170104():
    # check that we can find the right URL from an event dataset for
    # a GPS time that doesn't overlap with the event, and starts after
    # the end of the 32-second files (this used to not work)
    urls = locate.get_urls('L1', 1167558912.6, 1167559012.6, version=1)
    assert list(map(os.path.basename, urls)) == [
        "L-L1_GWOSC_4KHZ_R1-1167557889-4096.hdf5",
    ]
